#version 310 es
// .comp -> .glsl
precision highp float;

// NUM_X * NUM_Y * NUM_Z threads per work group.
// Max workgroup size per dimension : (32x32x1) < GL_MAX_COMPUTE_WORKGROUP_INVOCATIONS
layout(local_size_x=8, local_size_y=10)in;

layout(std430, binding=0)readonly buffer Input0{
    float data[];
}input0;

layout(std430, binding=1)readonly buffer Input1{
    float data[];
}input1;

layout(std430, binding=2)writeonly buffer Output{
    float data[];
}output0;

const float SIGN_RESHAPE=0.;
const float SIGN_BLOAT=1.;
const float SIGN_RECOVERY=2.;

const uint SIGN_RETRIEVAL_LOW=600u;
const uint SIGN_RETRIEVAL=1200u;

uniform float type;
uniform highp vec2 startPoint;
uniform highp vec2 endPoint;
uniform highp vec2 inputSize;
uniform float radius;
uniform float bloatScale;
uniform float isLowDevice;

vec2 warpReshape(vec2 point, vec2 start, vec2 end, float r, vec2 size){
    vec2 position=vec2(point.x, point.y);
    if (position.x!=-1.&&position.x!=1.&&position.y!=-1.&&position.y!=1.){
        vec2 realPosition=vec2(0.);
        realPosition.x=(point.x*.5+.5)*size.x;
        realPosition.y=(point.y*.5+.5)*size.y;

        float moveLength=sqrt(pow(end.x-start.x, 2.)+pow(end.y-start.y, 2.));
        float maxLength=r/2.;

        vec2 realEndPoint=vec2(end.x, end.y);

        if (moveLength>maxLength){
            realEndPoint.x=start.x+(end.x-start.x)*maxLength/moveLength;
            realEndPoint.y=start.y+(end.y-start.y)*maxLength/moveLength;
        }

        float currentLength=sqrt(pow(realPosition.x-start.x, 2.)+pow(realPosition.y-start.y, 2.));
        if (currentLength<r){
            float scale=(.6*pow(1.-(currentLength*currentLength)/(r*r), 6.));
            vec2 offset=vec2((realEndPoint.x-start.x)*scale, (realEndPoint.y-start.y)*scale);

            realPosition.x=realPosition.x+offset.x;
            realPosition.y=realPosition.y+offset.y;

            position.x=(realPosition.x/size.x-.5)*2.;
            position.y=(realPosition.y/size.y-.5)*2.;
        }
    }
    return position;
}

vec2 warpBloat(vec2 point, vec2 center, float r, float scale, vec2 size){
    vec2 position=vec2(point.x, point.y);
    if (position.x!=-1.&&position.x!=1.&&position.y!=-1.&&position.y!=1.){

        position.x=(position.x*.5+.5);
        position.y=1.-(position.y*.5+.5);

        vec2 tPosition=vec2(position.x, (position.y-center.y)*size.y/size.x+center.y);
        float tLength=sqrt(pow(tPosition.x-center.x, 2.)+pow(tPosition.y-center.y, 2.));

        if (tLength!=0.&&tLength<r){
            position.x=position.x-center.x;
            position.y=position.y-center.y;
            float percent=1.;
            if (scale<0.){
                percent=1.+((r-tLength)/r)*scale;
                percent=percent*percent;
                position.x=position.x*percent;
                position.y=position.y*percent;
            } else {
                percent=1.-(tLength/r)*scale;
                percent=percent*percent;
                vec2 p=vec2(position.x*r/tLength, position.y*r/tLength);
                vec2 off=vec2(p.x-position.x, p.y-position.y);
                off.x=off.x*percent;
                off.y=off.y*percent;
                position.x=p.x-off.x;
                position.y=p.y-off.y;
            }
            position.x=position.x+center.x;
            position.y=position.y+center.y;
        }

        position.x=(position.x-.5)*2.;
        position.y=(1.-position.y-.5)*2.;
    }
    return position;
}

vec2 warpRecovery(vec2 pointOriginal, vec2 pointInput, vec2 center, float r, float scale, vec2 size){
    vec2 position=vec2(pointInput.x, pointInput.y);
    if (position.x!=-1.&&position.x!=1.&&position.y!=-1.&&position.y!=1.){
        vec2 realPosition=vec2(0.);
        realPosition.x=(position.x*.5+.5)*size.x;
        realPosition.y=(position.y*.5+.5)*size.y;

        float currentLength=sqrt(pow(realPosition.x-center.x, 2.)+pow(realPosition.y-center.y, 2.));
        if (currentLength<r){
            vec2 originalPosition=vec2(pointOriginal.x, pointOriginal.y);
            vec2 originalRealPosition=vec2((originalPosition.x*.5+.5)*size.x, (originalPosition.y*.5+.5)*size.y);
            vec2 offset=vec2(realPosition.x-originalRealPosition.x, realPosition.y-originalRealPosition.y);
            float scale=.1*pow(1.-(currentLength*currentLength)/(r*r), 6.);
            realPosition.x=realPosition.x-offset.x*scale;
            realPosition.y=realPosition.y-offset.y*scale;

            position.x=(realPosition.x/size.x-.5)*2.;
            position.y=(realPosition.y/size.y-.5)*2.;
        }
    }
    return position;
}

void main()
{
    // 阻断操作，作用是等在当前执行范围(即设置的local_size_x和local_size_y)内的所有像素都执行完上一步操作，才往下执行。如果范围设置的是1x1 是不需要的
    // barrier();

    uint idx=gl_GlobalInvocationID.x;
    uint idy=gl_GlobalInvocationID.y;

    barrier();

    uint index = 0u;
    if (isLowDevice > 0.) {
        index=idy*SIGN_RETRIEVAL_LOW+idx;
    } else {
        index=idy*SIGN_RETRIEVAL+idx;
    }

    if (mod(float(index), 2.)==0.){
        float x=input0.data[index];
        float y=input0.data[index+1u];
        vec2 reshape=vec2(0.);
        if (type==SIGN_RESHAPE){
            reshape=warpReshape(vec2(x, y), startPoint, endPoint, radius, inputSize);
        } else if (type==SIGN_BLOAT) {
            reshape=warpBloat(vec2(x, y), endPoint, radius, bloatScale, inputSize);
        } else if (type==SIGN_RECOVERY) {
            float oriX=input1.data[index];
            float oriY=input1.data[index+1u];
            reshape=warpRecovery(vec2(oriX, oriY), vec2(x, y), endPoint, radius, bloatScale, inputSize);
        }
        output0.data[index]=reshape.x;
        output0.data[index+1u]=reshape.y;
    }
}